"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DynamicState = void 0;
const util_1 = require("@arktype/util");
const operand_js_1 = require("../shift/operand/operand.js");
const operator_js_1 = require("../shift/operator/operator.js");
const scanner_js_1 = require("../shift/scanner.js");
const string_js_1 = require("../string.js");
const shared_js_1 = require("./shared.js");
class DynamicState {
    ctx;
    scanner;
    // set root type to `any` so that all constraints can be applied
    root;
    branches = {
        prefixes: [],
        leftBound: null,
        intersection: null,
        union: null
    };
    finalizer;
    groups = [];
    constructor(def, ctx) {
        this.ctx = ctx;
        this.scanner = new scanner_js_1.Scanner(def);
    }
    error(message) {
        return (0, util_1.throwParseError)(message);
    }
    hasRoot() {
        return this.root !== undefined;
    }
    setRoot(root) {
        this.root = root;
    }
    unsetRoot() {
        const value = this.root;
        this.root = undefined;
        return value;
    }
    constrainRoot(...args) {
        this.root = this.root.constrain(args[0], args[1]);
    }
    finalize(finalizer) {
        if (this.groups.length)
            return this.error((0, shared_js_1.writeUnclosedGroupMessage)(")"));
        this.finalizeBranches();
        this.finalizer = finalizer;
    }
    reduceLeftBound(limit, comparator) {
        const invertedComparator = shared_js_1.invertedComparators[comparator];
        if (!(0, util_1.isKeyOf)(invertedComparator, shared_js_1.minComparators))
            return this.error((0, shared_js_1.writeUnpairableComparatorMessage)(comparator));
        if (this.branches.leftBound) {
            return this.error((0, shared_js_1.writeMultipleLeftBoundsMessage)(this.branches.leftBound.limit, this.branches.leftBound.comparator, limit, invertedComparator));
        }
        this.branches.leftBound = {
            comparator: invertedComparator,
            limit
        };
    }
    finalizeBranches() {
        this.assertRangeUnset();
        if (this.branches.union) {
            this.pushRootToBranch("|");
            this.root = this.branches.union;
        }
        else if (this.branches.intersection) {
            this.pushRootToBranch("&");
            this.root = this.branches.intersection;
        }
        else
            this.applyPrefixes();
    }
    finalizeGroup() {
        this.finalizeBranches();
        const topBranchState = this.groups.pop();
        if (!topBranchState)
            return this.error((0, shared_js_1.writeUnmatchedGroupCloseMessage)(this.scanner.unscanned));
        this.branches = topBranchState;
    }
    addPrefix(prefix) {
        this.branches.prefixes.push(prefix);
    }
    applyPrefixes() {
        while (this.branches.prefixes.length) {
            const lastPrefix = this.branches.prefixes.pop();
            this.root =
                lastPrefix === "keyof" ?
                    this.root.keyof()
                    : (0, util_1.throwInternalError)(`Unexpected prefix '${lastPrefix}'`);
        }
    }
    pushRootToBranch(token) {
        this.assertRangeUnset();
        this.applyPrefixes();
        const root = this.root;
        this.branches.intersection = this.branches.intersection?.and(root) ?? root;
        if (token === "|") {
            this.branches.union =
                this.branches.union?.or(this.branches.intersection) ??
                    this.branches.intersection;
            this.branches.intersection = null;
        }
        this.root = undefined;
    }
    parseUntilFinalizer() {
        return (0, string_js_1.parseUntilFinalizer)(new DynamicState(this.scanner.unscanned, this.ctx));
    }
    parseOperator() {
        return (0, operator_js_1.parseOperator)(this);
    }
    parseOperand() {
        return (0, operand_js_1.parseOperand)(this);
    }
    assertRangeUnset() {
        if (this.branches.leftBound) {
            return this.error((0, shared_js_1.writeOpenRangeMessage)(this.branches.leftBound.limit, this.branches.leftBound.comparator));
        }
    }
    reduceGroupOpen() {
        this.groups.push(this.branches);
        this.branches = {
            prefixes: [],
            leftBound: null,
            union: null,
            intersection: null
        };
    }
    previousOperator() {
        return (this.branches.leftBound?.comparator ??
            this.branches.prefixes.at(-1) ??
            (this.branches.intersection ? "&"
                : this.branches.union ? "|"
                    : undefined));
    }
    shiftedByOne() {
        this.scanner.shift();
        return this;
    }
}
exports.DynamicState = DynamicState;
