"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeInvalidLimitMessage = exports.parseRightBound = exports.singleEqualsMessage = exports.getBoundKinds = exports.writeIncompatibleRangeMessage = exports.comparatorStartChars = exports.parseBound = void 0;
const schema_1 = require("@arktype/schema");
const util_1 = require("@arktype/util");
const shared_js_1 = require("../../reduce/shared.js");
const date_js_1 = require("../operand/date.js");
const parseBound = (s, start) => {
    const comparator = shiftComparator(s, start);
    if (s.root.hasKind("unit")) {
        if (typeof s.root.unit === "number") {
            s.reduceLeftBound(s.root.unit, comparator);
            s.unsetRoot();
            return;
        }
        if (s.root.unit instanceof Date) {
            const literal = `d'${s.root.description ?? s.root.unit.toISOString()}'`;
            s.unsetRoot();
            s.reduceLeftBound(literal, comparator);
            return;
        }
    }
    return (0, exports.parseRightBound)(s, comparator);
};
exports.parseBound = parseBound;
const oneCharComparators = {
    "<": true,
    ">": true
};
exports.comparatorStartChars = {
    "<": 1,
    ">": 1,
    "=": 1
};
const shiftComparator = (s, start) => s.scanner.lookaheadIs("=") ? `${start}${s.scanner.shift()}`
    : (0, util_1.isKeyOf)(start, oneCharComparators) ? start
        : s.error(exports.singleEqualsMessage);
const writeIncompatibleRangeMessage = (l, r) => `Bound kinds ${l} and ${r} are incompatible`;
exports.writeIncompatibleRangeMessage = writeIncompatibleRangeMessage;
const getBoundKinds = (comparator, limit, root, boundKind) => {
    if (root.extends(schema_1.tsKeywords.number)) {
        if (typeof limit !== "number") {
            return (0, util_1.throwParseError)((0, exports.writeInvalidLimitMessage)(comparator, limit, boundKind));
        }
        return (comparator === "==" ? ["min", "max"]
            : comparator[0] === ">" ? ["min"]
                : ["max"]);
    }
    if (root.extends(schema_1.internalKeywords.lengthBoundable)) {
        if (typeof limit !== "number") {
            return (0, util_1.throwParseError)((0, exports.writeInvalidLimitMessage)(comparator, limit, boundKind));
        }
        return (comparator === "==" ? ["minLength", "maxLength"]
            : comparator[0] === ">" ? ["minLength"]
                : ["maxLength"]);
    }
    if (root.extends(schema_1.jsObjects.Date)) {
        // allow either numeric or date limits
        return (comparator === "==" ? ["after", "before"]
            : comparator[0] === ">" ? ["after"]
                : ["before"]);
    }
    return (0, util_1.throwParseError)((0, schema_1.writeUnboundableMessage)(root.expression));
};
exports.getBoundKinds = getBoundKinds;
exports.singleEqualsMessage = "= is not a valid comparator. Use == to check for equality";
const openLeftBoundToRoot = (leftBound) => ({
    rule: (0, date_js_1.isDateLiteral)(leftBound.limit) ?
        (0, date_js_1.extractDateLiteralSource)(leftBound.limit)
        : leftBound.limit,
    exclusive: leftBound.comparator.length === 1
});
const parseRightBound = (s, comparator) => {
    // store the node that will be bounded
    const previousRoot = s.unsetRoot();
    const previousScannerIndex = s.scanner.location;
    s.parseOperand();
    const limitNode = s.unsetRoot();
    // after parsing the next operand, use the locations to get the
    // token from which it was parsed
    const limitToken = s.scanner.sliceChars(previousScannerIndex, s.scanner.location);
    s.root = previousRoot;
    if (!limitNode.hasKind("unit") ||
        (typeof limitNode.unit !== "number" && !(limitNode.unit instanceof Date)))
        return s.error((0, exports.writeInvalidLimitMessage)(comparator, limitToken, "right"));
    const limit = limitNode.unit;
    // apply the newly-parsed right bound
    const exclusive = comparator.length === 1;
    // if the comparator is ==, both the min and max of that pair will be applied
    for (const kind of (0, exports.getBoundKinds)(comparator, typeof limit === "number" ? limit : limitToken, previousRoot, "right"))
        s.constrainRoot(kind, { rule: limit, exclusive });
    if (!s.branches.leftBound)
        return;
    // if there's an open left bound, perform additional validation and apply it
    if (!(0, util_1.isKeyOf)(comparator, shared_js_1.maxComparators))
        return s.error((0, shared_js_1.writeUnpairableComparatorMessage)(comparator));
    const lowerBoundKind = (0, exports.getBoundKinds)(s.branches.leftBound.comparator, s.branches.leftBound.limit, previousRoot, "left");
    s.constrainRoot(lowerBoundKind[0], openLeftBoundToRoot(s.branches.leftBound));
    s.branches.leftBound = null;
};
exports.parseRightBound = parseRightBound;
const writeInvalidLimitMessage = (comparator, limit, boundKind) => `Comparator ${boundKind === "left" ? shared_js_1.invertedComparators[comparator] : comparator} must be ${boundKind === "left" ? "preceded" : "followed"} by a corresponding literal (was ${limit})`;
exports.writeInvalidLimitMessage = writeInvalidLimitMessage;
